C******************************
C  BEGINING OF HELMHOLTZ SOLVER:
C******************************
 
      SUBROUTINE HELMHOLT
     &    (G,F,SOLUTN,ZK2,NX1,NY1,M,H1,NU1,NU2,NCYC,NX,MY)
*********************************************************************
*********************************************************************
**   Solves helmholtz equation:                                    **
**   (\nabla^2 - ZK2) SOLUTN=F, with dirichlet                      **
** b.c.  The b.c and the initial guess for the solution are given  **
** in the array G.                                                 **
**                                                                 **
**   The dimension of the arrays should be:                        **
**    (nx,my)= ( p*2^(M-1)+1, q*2^(M-1)+1 ).                       **
** p and q are small integers, equal to the num of grid boxes in   **
** the coarsest grid used in the multi grid solution.              **
**                                                                 **
** Note that num of boxes on one side of the grid =                **
**     = num of grid points on the side - 1.                       **
**                                                                 **
** In the prog p=NX1; q=NX2.                                       **
** M is the finest level of the multigrid. M=1 means use (p,q)     **
** boxes; M=2 means use twice as many boxes on each lide of grid.  **
**                                                                 **
** H1 is the delx for the coasest grid. (NU1,NU2) should be (1,2). **
** NCYCL should be 3-5 for a bad initial guess, and 1 for a good   **
** initial guess.                                                  **
**   The dimension of the work array Q should be NX*NY*3.          **
**                                                                 **
*********************************************************************
*********************************************************************
C
      IMPLICIT REAL*8  (A-H,O-Z)
      DIMENSION G(NX,MY),F(NX,MY),SOLUTN(NX,MY)
C   CHANGE NDIMQ IN ALL SUBROUTINES OF SOLVER!!
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      COMMON/CGRDIQ/ IQ
      DIMENSION IST(200)
C
C  CHECK THAT DIMENSION OF Q IS LARGE ENOUGH FOR SPECIFIED N,M:
      IF (NDIMQ.LT.NX*MY*3)
     1   STOP' POISS1: Dimension of array Q too small.'
C
C  CHECK THAT SOR is used for low froud numbers:
c$$$      IF (ZK2.LT.1000.) THEN
c$$$         WRITE(*,*) '************************************************* '
c$$$         WRITE(*,*) '** POISS1: Small FROUD, use RELAX, not RELAXZ! ** '
c$$$         WRITE(*,*) '************************************************* '
c$$$         STOP
c$$$      END IF
C
!      IF (NCYC.LT.10) NCYC=10
      IQ=1
      DO 1 K=1,M
      K2=2**(K-1)
      CALL GRDFN(K,NX1*K2+1,NY1*K2+1,H1/K2)
  1   CALL GRDFN(K+M,NX1*K2+1,NY1*K2+1,H1/K2)
      WU=0.0D0
C
      CALL PUTF1(M,G,0,NX,MY)
      CALL PUTF1(2*M,F,2,NX,MY)
      DO 5 IC=1,NCYC
      DO 3 KM=1,M
      K=1+M-KM
      IF (K.NE.M) CALL PUTZ(K)
      DO 2 IR=1,NU1
    2 CALL RELAX(K,K+M,WU,M,ZK2)
    3 IF (K.GT.1) CALL RESCAL(K,K+M,K+M-1,ZK2)
      DO 5 K=1,M
      DO 4 IR=1,NU2
    4 CALL RELAX(K,K+M,WU,M,ZK2)
    5 IF (K.LT.M) CALL INTADD(K,K+1)
      CALL GETSOL(M,SOLUTN,NX,MY)
      RETURN
      END
C
C
      SUBROUTINE GRDFN(K,M,N,HH)
      IMPLICIT REAL*8  (A-H,O-Z)
      COMMON/GRD/NST(20),IMX(20),JMX(20),H(20)
      COMMON/CGRDIQ/ IQ
C     DATA IQ/1/
      NST(K)=IQ
      IMX(K)=M
      JMX(K)=N
      H(K)=HH
      IQ=IQ+M*N
      RETURN
      END
C
      SUBROUTINE KEY(K,IST,M,N,HH)
      IMPLICIT REAL*8  (A-H,O-Z)
      COMMON/GRD/NST(20),IMX(20),JMX(20),H(20)
      DIMENSION IST(200)
      M=IMX(K)
      N=JMX(K)
      IS=NST(K)-N-1
      DO 1 I=1,M
      IS=IS + N
    1 IST(I)=IS
      HH=H(K)
      RETURN
      END
C
      SUBROUTINE PUTF1(K,F,NH,NX,MY)
      IMPLICIT REAL*8  (A-H,O-Z)
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      DIMENSION IST(200),F(NX,MY)
      CALL KEY (K,IST,II,JJ,H)
      H2=H**NH
      DO 1 I=1,II
       DO 1 J=1,JJ
    1 Q(IST(I)+J)=F(I,J)*H2
      RETURN
      END
C
      SUBROUTINE GETSOL(K,SOLUTN,NX,MY)
      IMPLICIT REAL*8  (A-H,O-Z)
      DIMENSION SOLUTN(NX,MY)
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      DIMENSION IST(200)
      CALL KEY(K,IST,II,JJ,H)
      DO 1 I=1,II
      DO 1 J=1,JJ
    1 SOLUTN(I,J)=Q(IST(I)+J)
      RETURN
      END
C
C
      SUBROUTINE PUTZ(K)
      IMPLICIT REAL*8  (A-H,O-Z)
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      DIMENSION IST(200)
      CALL KEY(K,IST,II,JJ,H)
      DO 1 I=1,II
      DO 1 J=1,JJ
    1 Q(IST(I)+J)=0.0D0
      RETURN
      END
C
      SUBROUTINE RELAX(K,KRHS,WU,M,ZK2)
      IMPLICIT REAL*8  (A-H,O-Z)
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      DIMENSION IST(200),IRHS(200)
      CALL KEY(K,IST,II,JJ,H)
      CALL KEY(KRHS,IRHS,II,JJ,H)
      DIAG=4.0D0+ZK2*H**2
      I1=II-1
      J1=JJ-1
c      ERR=0.0D0
      DO 1 I=2,I1
         IR=IRHS(I)
         IQ=IST(I)
         IM=IST(I-1)
         IP=IST(I+1)
         DO 1 J=2,J1
            A=Q(IR+J)-Q(IQ+J+1)-Q(IQ+J-1)-Q(IM+J)-Q(IP+J)
c            ERR=ERR+(A+DIAG*Q(IQ+J))**2
 1          Q(IQ+J)=-A/DIAG
c      ERR=SQRT(ERR)/H
      WU=WU+4.0D0**(K-M)
C     WRITE( 6,2) K,ERR,WU
C   2 FORMAT('  LEVEL',I2,'   RESIDUAL NORM=',1PE10.3,'   WORK=',0PF7.3)
      RETURN
      END
C
      SUBROUTINE INTADD(KC,KF)
      IMPLICIT REAL*8  (A-H,O-Z)
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      DIMENSION ISTC(200),ISTF(200)
      CALL KEY(KC,ISTC,IIC,JJC,HC)
      CALL KEY(KF,ISTF,IIF,JJF,HF)
      DO 1 IC=2,IIC
        IF=2*IC-1
        JF=1
        IFO=ISTF(IF)
        IFM=ISTF(IF-1)
        ICO=ISTC(IC)
        ICM=ISTC(IC-1)
        DO 1 JC=2,JJC
          JF=JF+2
          A=0.5D0*(Q(ICO+JC)+Q(ICO+JC-1))
          AM=0.5D0*(Q(ICM+JC)+Q(ICM+JC-1))
          Q(IFO+JF) = Q(IFO+JF)+Q(ICO+JC)
          Q(IFM+JF) = Q(IFM+JF)+0.5D0*(Q(ICO+JC)+Q(ICM+JC))
          Q(IFO+JF-1)=Q(IFO+JF-1)+A
 1      Q(IFM+JF-1) = Q(IFM+JF-1)+0.5D0*(A+AM)
      RETURN
      END
C
      SUBROUTINE RESCAL(KF,KRF,KRC,ZK2)
      IMPLICIT REAL*8  (A-H,O-Z)
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      DIMENSION IUF(200),IRF(200),IRC(200)
      CALL KEY(KF,IUF,IIF,JJF,HF)
      CALL KEY(KRF,IRF,IIF,JJF,HF)
      CALL KEY(KRC,IRC,IIC,JJC,HC)
      DIAG=4.0D0+ZK2*HF**2
      IIC1=IIC-1
      JJC1=JJC-1
      DO 1 IC=2,IIC1
      ICR=IRC(IC)
      IF=2*IC-1
      JF=1
      IFR=IRF(IF)
      IF0=IUF(IF)
      IFM=IUF(IF-1)
      IFP=IUF(IF+1)
      DO 1 JC=2,JJC1
      JF=JF+2
      S=Q(IF0+JF+1)+Q(IF0+JF-1)+Q(IFM+JF)+Q(IFP+JF)
    1 Q(ICR+JC)=4.0D0*(Q(IFR+JF)-S+DIAG*Q(IF0+JF))
      RETURN
      END
C
      SUBROUTINE INTRP3(KC,KF)
      IMPLICIT REAL*8  (A-H,O-Z)
      PARAMETER(NDIMQ=1000000)
      COMMON /CPOIS1/ Q(NDIMQ)
      DIMENSION IUF(200),IUC(400)
      CALL KEY(KC,IUC,IIC,JJC,HC)
      CALL KEY(KF,IUF,IIF,JJF,HF)
      DO 20 IC=1,IIC
      IC0 = IUC(IC)
      IF0 = IUF(2*IC-1)
      Q(IF0+1) = Q(IC0+1)
      Q(IF0+2) = (5*Q(IC0+1)+15*Q(IC0+2)-5*Q(IC0+3)+Q(IC0+4))/16
      JJC2 = JJC-2
      DO 10 JC=2,JJC2
      JF = 2*JC-1
      Q(IF0+JF) = Q(IC0+JC)
      Q(IF0+JF+1) = (-Q(IC0+JC-1)+9.0D0*Q(IC0+JC)
     +               +9.0D0*Q(IC0+JC+1)-Q(IC0+JC+2))/16.0D0
   10 CONTINUE
      Q(IF0+JJF-2) = Q(IC0+JJC-1)
      Q(IF0+JJF-1) = (Q(IC0+JJC-3)-5.0D0*Q(IC0+JJC-2)
     +                +15.0D0*Q(IC0+JJC-1)+5.0D0*Q(IC0+JJC))/16.0D0
      Q(IF0+JJF) = Q(IC0+JJC)
   20 CONTINUE
      IM1 = IUF(1)
      I0 = IUF(2)
      IP1 = IUF(3)
      IP3 = IUF(5)
      IP5 = IUF(7)
      DO 30 J=1,JJF
      Q(I0+J) = (5.0D0*Q(IM1+J)+15.0D0*Q(IP1+J)
     +           -5.0D0*Q(IP3+J)+Q(IP5+J))/16.0D0
   30 CONTINUE
      IIF3 = IIF-3
      DO 40 I=4,IIF3,2
      IM3 = IUF(I-3)
      IM1 = IUF(I-1)
      I0  = IUF(I)
      IP1 = IUF(I+1)
      IP3 = IUF(I+3)
      DO 40 J=1,JJF
      Q(I0+J) = (-Q(IM3+J)+9.0D0*Q(IM1+J)
     +           +9.0D0*Q(IP1+J)-Q(IP3+J))/16.0D0
   40 CONTINUE
      IM5 = IUF(IIF-6)
      IM3 = IUF(IIF-4)
      IM1 = IUF(IIF-2)
      I0  = IUF(IIF-1)
      IP1 = IUF(IIF)
      DO 50 J=1,JJF
      Q(I0+J) = (Q(IM5+J)-5.0D0*Q(IM3+J)
     +           +15.0D0*Q(IM1+J)+5.0D0*Q(IP1+J))/16.0D0
   50 CONTINUE
      RETURN
      END

C******************************
C       END OF HELMHOLTZ SOLVER:
C******************************
